/* 
 * Licensed Materials - Property of IBM and/or HCL 
 *  Copyright IBM Corporation 2005, 2016. All Rights Reserved. 
 *  Copyright HCL Technologies Ltd. 2017,2023. All Rights Reserved. 
 *  
 * Note to U.S. Government Users Restricted Rights: 
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *
 */
#include "XMLDHI.hpp"
#include <xercesc/framework/LocalFileInputSource.hpp>
#include <xercesc/framework/MemBufInputSource.hpp>
#include <xercesc/framework/URLInputSource.hpp>
#include <xercesc/sax/HandlerBase.hpp>
#include <xercesc/sax/SAXParseException.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/TransService.hpp>
#include <xercesc/util/XMLURL.hpp>

XERCES_CPP_NAMESPACE_USE

/////////////////////////////////////////////////////////////////////////////
// Helper classes

/////////////////////////////////////////////////////////////////////////////
// ErrorHandler
//
class ErrorHandlerImpl : public HandlerBase
{
public:
  ErrorHandlerImpl(char* errorMessage)
  {
    fErrorMessage = errorMessage;
    memset(fErrorMessage, 0, strlen(errorMessage));
  }

  ~ErrorHandlerImpl()
  {
  }

public:
  char* getErrorMessage()
  {
    return fErrorMessage;
  }

public:
  void error(const SAXParseException& ex)
  {
    sprintf(
        fErrorMessage
      , "%s\nSAXParseException: Error in %s at line %i, column %i.\n%s\n"
      , fErrorMessage
      , XMLString::transcode(ex.getSystemId())
      , ex.getLineNumber()
      , ex.getColumnNumber()
      , XMLString::transcode(ex.getMessage()));
  }
  
  void fatalError(const SAXParseException& ex)
  {
    sprintf(
        fErrorMessage
      , "%s\nSAXParseException: Fatal error in %s at line %i, column %i.\n%s\n"
      , fErrorMessage
      , XMLString::transcode(ex.getSystemId())
      , ex.getLineNumber()
      , ex.getColumnNumber()
      , XMLString::transcode(ex.getMessage()));
  }

  void warning(const SAXParseException& ex)
  {
    sprintf(
        fErrorMessage
      , "%s\nSAXParseException: Warning in %s at line %i, column %i.\n%s\n"
      , fErrorMessage
      , XMLString::transcode(ex.getSystemId())
      , ex.getLineNumber()
      , ex.getColumnNumber()
      , XMLString::transcode(ex.getMessage()));
  }

private:
  char* fErrorMessage;
};

/////////////////////////////////////////////////////////////////////////////
//
// Class XMLDocumentHandlerImpl

XMLDocumentHandlerImpl::XMLDocumentHandlerImpl() :
    fCurrentOffset(0)
  , fMaxSize(0)
  , fCount(0)
  , fSourceFile(0)
  , fBundleFile(0)
  , fSourceFileName(0)
  , fHeaderBuff(0)
  , fHeaderBuffSize(0)
  , fTrailerBuff(0)
  , fTrailerBuffSize(0)
  , fContentBuff(0)
  , fParser(0)
  , fPayloadHasStarted(false)
{
}

XMLDocumentHandlerImpl::~XMLDocumentHandlerImpl()
{
  if (fEncoding)
    XMLString::release(&fEncoding);
}

// getInputSource ///////////////////////////////////////////////////////////

InputSource* XMLDocumentHandlerImpl::getInputSource(const char* url)
{
  fSourceFileName = XMLString::transcode(url);
  fSourceFile = XMLPlatformUtils::openFile(fSourceFileName);
  
  XMLURL xmlUrl;
  if (XMLURL::parse(fSourceFileName, xmlUrl))
  {
    if (xmlUrl.isRelative())
      return new LocalFileInputSource(fSourceFileName);
    else
      return new URLInputSource(xmlUrl);
  }
  else
    return new LocalFileInputSource(fSourceFileName);
}

// getFileName //////////////////////////////////////////////////////////////

void XMLDocumentHandlerImpl::getFileName(char* fileName /*OUT*/)
{
#ifndef __MVS__
  XMLCh fileNameWOExt[MAX_PATH] = {chNull};
  XMLString::subString(
      fileNameWOExt
    , fSourceFileName
    , 0
    , XMLString::lastIndexOf(fSourceFileName, chPeriod));
  sprintf(
      fileName
    , "%s_%03d.xml"
    , XMLString::transcode(fileNameWOExt)
    , ++fCount);
#else
  sprintf(
      fileName
    , "DD:XML%3.3d"
    , ++fCount);
#endif
}

// writeFile ////////////////////////////////////////////////////////////////

void XMLDocumentHandlerImpl::writeFile(unsigned int toWrite, bool writeHeader, bool writeTrailer)
{
  // Build the file name
  char sourceFileName[MAX_PATH] = {'\0'};
  getFileName(sourceFileName);

  // Create a new file
  FileHandle sourceFile = XMLPlatformUtils::openFileToWrite(sourceFileName);
  if (sourceFile)
  {
    // Write the header
    if (writeHeader)
    {
      XMLPlatformUtils::writeBufferToFile(
          sourceFile
        , fHeaderBuffSize
        , fHeaderBuff);
    }
    // Read the input file and write its content into the output file using pages
    // of PAGE_SIZE size
    while (1)
    {
      memset(fContentBuff, 0, PAGE_SIZE);
      if (toWrite > PAGE_SIZE)
      {
        if (XMLPlatformUtils::readFileBuffer(fSourceFile, PAGE_SIZE, fContentBuff))
        {
          XMLPlatformUtils::writeBufferToFile(sourceFile, PAGE_SIZE, fContentBuff);
        }
        toWrite -= PAGE_SIZE;
      }
      else
      {
        if (XMLPlatformUtils::readFileBuffer(fSourceFile, toWrite, fContentBuff))
        {
          XMLPlatformUtils::writeBufferToFile(sourceFile, toWrite, fContentBuff);
        }
        break;
      }
    }
    // Write the trailer
    if (writeTrailer)
    {
      XMLPlatformUtils::writeBufferToFile(
          sourceFile
        , fTrailerBuffSize
        , fTrailerBuff);
    }
    XMLPlatformUtils::closeFile(sourceFile);
  }
}

// parse ////////////////////////////////////////////////////////////////////

unsigned int XMLDocumentHandlerImpl::parse(
    const char* xmlUrl
  , unsigned int maxSize
  , char* errMsg)
{
  InputSource* inputSource = 0;
  fMaxSize = maxSize;
  try
  {
    // Create the SAX Parser
    fParser = new SAXParser(
        0
      , XMLPlatformUtils::fgMemoryManager
      , 0);
    if (!fParser)
      throw;

    // Set the parser features
    fParser->installAdvDocHandler(this);
    fParser->setErrorHandler(new ErrorHandlerImpl(errMsg));

    fParser->setValidationConstraintFatal(true);
    fParser->setValidationScheme(fParser->Val_Never);
    fParser->setDoSchema(false);
    fParser->setDoNamespaces(true);
    fParser->setValidationSchemaFullChecking(false);
    fParser->setCalculateSrcOfs(true);

    // Create the input source
    inputSource = getInputSource(xmlUrl);
		if (!inputSource)
			throw;

    // Parse the input
    fParser->parse(*inputSource);
  }
  catch (SAXParseException& ex)
  {
    sprintf(
        errMsg
      , "SAXParseException: Error in %s, at line %i, column %i\n%s\n"
      , XMLString::transcode(ex.getSystemId())
      , ex.getLineNumber()
      , ex.getColumnNumber()
      , XMLString::transcode(ex.getMessage()));
  }
  catch (XMLException& ex)
  {
    sprintf(
        errMsg
      , "XMLException: Error in %s, at line %i\n%s\n"
      , XMLString::transcode(ex.getSrcFile())
      , ex.getSrcLine()
      , XMLString::transcode(ex.getMessage()));
  }

  // Clean up
  if (fParser)
    delete fParser;
  if (inputSource)
    delete inputSource;
  if (fContentBuff)
    delete [] fContentBuff;
  if (fHeaderBuff)
    delete [] fHeaderBuff;
  if (fTrailerBuff)
    delete [] fTrailerBuff;
  if (fSourceFile)
    XMLPlatformUtils::closeFile(fSourceFile);

  fprintf(stderr, "Resulting number of files: %d\n", fCount);
  return fCount;
}

/////////////////////////////////////////////////////////////////////////////
//
// XMLDocumentHandler interface implementation

void XMLDocumentHandlerImpl::docCharacters(
    const XMLCh* const chars
  , const unsigned int length
  , const bool cdataSection)
{
}

void XMLDocumentHandlerImpl::docComment(
  const XMLCh* const comment)
{
}

void XMLDocumentHandlerImpl::docPI(
    const XMLCh* const target
  , const XMLCh* const data)
{
}

// endDocument //////////////////////////////////////////////////////////////

void XMLDocumentHandlerImpl::endDocument()
{
  int64_t offset = fParser->getSrcOffset();
  if (fCurrentOffset > MAX_UINT)
    offset = offset + MAX_UINT + 1;

  writeFile((unsigned int)(offset - fCurrentOffset), true, false);
  return;
}

// endElement ///////////////////////////////////////////////////////////////

void XMLDocumentHandlerImpl::endElement(
    const XMLElementDecl& elemDecl
  , const unsigned int uriId
  , const bool isRoot
  , const XMLCh* const prefixName)
{
  const XMLCh* name = elemDecl.getElementName()->getLocalPart();

  // Check if the element is among those of interest
  if (checkElementName(name))
  {
    int64_t offset = fParser->getSrcOffset();
    if (fCurrentOffset > offset || fCurrentOffset > MAX_UINT)
      offset = offset + MAX_UINT + 1;

    unsigned int toRead = (unsigned int)(offset - fCurrentOffset);
    if (fMaxSize == ANY_SIZE ||
        offset > (fMaxSize * (fCount + 1)))
    {
      if (!fContentBuff)
      {
        fContentBuff = new XMLByte[PAGE_SIZE];
        memset(fContentBuff, 0, PAGE_SIZE);
      }

      // Save the content into a new file
      writeFile(toRead);
      fCurrentOffset = offset;
    }
  }
}

void XMLDocumentHandlerImpl::endEntityReference(
    const XMLEntityDecl& entDecl)
{
}

void XMLDocumentHandlerImpl::ignorableWhitespace(
    const XMLCh* const chars
  , const unsigned int length
  , const bool cdataSection)
{
}

void XMLDocumentHandlerImpl::resetDocument()
{
}

void XMLDocumentHandlerImpl::startDocument()
{
}

// startElement /////////////////////////////////////////////////////////////

void XMLDocumentHandlerImpl::startElement(
    const XMLElementDecl& elemDecl
  , const unsigned int uriId
  , const XMLCh* const prefixName
  , const RefVectorOf<XMLAttr>& attrList
  , const unsigned int attrCount
  , const bool isEmpty
  , const bool isRoot)
{
  if (fPayloadHasStarted)
    return;

  const XMLCh* name = elemDecl.getElementName()->getLocalPart();
  if (isRoot)
  {
    XMLByte trailerBuff[512];

    // Create the content of the trailer buffer
    fTrailerBuffSize = 
        XMLString::stringLen(gOpenAngleSlash)
      + XMLString::stringLen(prefixName)
      + XMLString::stringLen(gColon)
      + XMLString::stringLen(name)
      + XMLString::stringLen(gCloseAngle);
    fTrailerBuff = new XMLByte[fTrailerBuffSize * 2];
    if (fTrailerBuff)
    {
      memset(trailerBuff, 0, 512);
      XMLString::catString((XMLCh*) trailerBuff, gOpenAngleSlash);
      XMLString::catString((XMLCh*) trailerBuff, prefixName);
      XMLString::catString((XMLCh*) trailerBuff, gColon);
      XMLString::catString((XMLCh*) trailerBuff, name);
      XMLString::catString((XMLCh*) trailerBuff, gCloseAngle);
      
      // Convert to the document's encoding
      XMLTransService::Codes result;
      XMLTranscoder* transcoder = XMLPlatformUtils::fgTransService->makeNewTranscoderFor(
          fEncoding
        , result
        , 64);
      if (transcoder)
      {
        memset(fTrailerBuff, 0, fTrailerBuffSize * 2);
        transcoder->transcodeTo(
            (XMLCh*) trailerBuff
          , fTrailerBuffSize
          , fTrailerBuff
          , fTrailerBuffSize * 2
          , fTrailerBuffSize
          , XMLTranscoder::UnRep_Throw);
        delete transcoder;
      }
    }
  }
  if (checkElementName(name))
  {
    // If this is the first element of interest, store the content of the input
    // file up to this offset into the header buffer
    fPayloadHasStarted = true;
    fHeaderBuff = new XMLByte[fHeaderBuffSize];
    if (fHeaderBuff)
    {
      memset(fHeaderBuff, 0, fHeaderBuffSize);
      XMLPlatformUtils::readFileBuffer(fSourceFile, fHeaderBuffSize, fHeaderBuff);
      fCurrentOffset = fHeaderBuffSize;
    }
  }
  else
  {
    fHeaderBuffSize = fParser->getSrcOffset();
  }
}

void XMLDocumentHandlerImpl::startEntityReference(
    const XMLEntityDecl& entDecl)
{
}

// XMLDecl //////////////////////////////////////////////////////////////////

void XMLDocumentHandlerImpl::XMLDecl(
    const XMLCh* const versionStr
  , const XMLCh* const encodingStr
  , const XMLCh* const standaloneStr
  , const XMLCh* const autoEncodingStr)
{
  if (encodingStr && *encodingStr)
    fEncoding = XMLString::replicate(encodingStr);
	else
    fEncoding = XMLString::replicate(autoEncodingStr);
}
